#include <map>
#include <cstdio>
#include <cstring>

using std::strchr;
using std::fprintf;
using std::sprintf;
using std::strtod;

#define ARGH_VERSION_2
#include "argh.hh"

using std::map;
using std::pair;

class argh_aliasmap : public map<string, int>
{
};
class argh_descsmap : public map<int, pair<string,string> >
{
};
class argh_atypemap : public map<int, ParamHandler::ArgInfo>
{
};

ParamHandler::ParamHandler()
{
    aliases  = new argh_aliasmap();
    descs    = new argh_descsmap();
    argtypes = new argh_atypemap();
}
ParamHandler::~ParamHandler()
{
    delete aliases;
    delete descs;
    delete argtypes;
}

int ParamHandler::ShortOpt(int key, const char *s)
{
    argh_atypemap::const_iterator i = argtypes->find(key);
    if(i == argtypes->end())
    {
        if(key=='=' && longo.size())
            fprintf(stderr, "%s: Parameter error: `--%s%s' does not take a value.\n",
                    A0, polarity?"":"no-", longo.c_str());
        else
            fprintf(stderr, "%s: illegal option -- %c\n", A0, key);
        return ParseError();
    }
    switch(i->second.type)
    {
        case 0: /* bool */
            shortpointer = s;
            return key;
        case 1: /* int */
        {
            int value = 0; bool nega = false;
            if(*s=='=')++s;
            if(!*s && (argpos+1) < argc)s = argv[++argpos];
            if(i->second.min < 0 && *s=='-') { nega = true; ++s; }
            if(s[0] < '0' || s[0] > '9')goto ArgMissing;
            while(s[0] >= '0' && s[0] <= '9')
                value = value*10 + (*s++ - '0');
            if(value < i->second.min || value > i->second.max)
            {
            	char Buf[128]; sprintf(Buf, "%d", value); param = Buf;
            	goto RangeFailure;
            }
            shortpointer = s;
            intparm = value;
            return key;
        }
        case 2: /* float */
        {
            if(*s=='=')++s;
            if(!*s && (argpos+1) < argc)s = argv[++argpos];
            char *newend;
        	doubleparm = strtod(s, &newend);
        	if(newend == s)goto ArgMissing;
        	if(doubleparm < i->second.fmin || doubleparm > i->second.fmax)
        	{
            	char Buf[256]; sprintf(Buf, "%g", doubleparm); param = Buf;
            	goto RangeFailure;
        	}
        	shortpointer = s;
            break;
        }
        case 3: /* string */
        {
            param = s;
            if(param.size() && param[0]=='=' && longobuf.size())param=s+1;
            else if(!param.size() && (argpos+1) < argc && (argv[argpos+1][0] != '-' || !argv[argpos+1][1]))
                param = argv[++argpos];
            else if(!param.size() && i->second.min > 0)
            {
ArgMissing:     if(longo.size())
                    fprintf(stderr, "Switch `--%s' requires an argument.\n", longo.c_str());
                else
                    fprintf(stderr, "Option `-%c' requires an argument.\n", key);
                return ParseError();
            }
            if((int)param.size() < i->second.min || param.size() > i->second.maxl)
            {
RangeFailure:   fprintf(stderr, "%s: parameter length/value out of range `%s'\n", A0, param.c_str());
                return ParseError();
            }
            shortpointer = "";
            return key;
        }
    }
    /* Internal error */
    fprintf(stderr, "%s: Internal error when parsing key '%d', s='%s'\n", A0, key, s);
    return key;
}

int ParamHandler::GetParam()
{
    if(error)return -1;
    
Retry:
	polarity = true;
	
    if(shortpointer && *shortpointer)
        return ShortOpt(*shortpointer, shortpointer+1);

    if(++argpos >= argc)return -1;
    
    /* new arg */
    const char *s = argv[argpos];
    if(!opts || *s!='-' || !s[1])
    {
        param = s;
        return 0;
    }
    
    longo = "";
    longobuf = "";
    
    if(s[1] != '-')
        return ShortOpt(s[1], s+2);

    if(!s[2]) /* -- */
    {
        opts = false;
        goto Retry;
    }
    
    /* --something */
    longo = s+2;
    bool nega = false, negafail = false;
    if(longo.substr(0, 4) == "not-")          { nega=true; longo.erase(0, 4); }
    else if(longo.substr(0, 3) == "no-")      { nega=true; longo.erase(0, 3); }
    else if(longo.substr(0, 8) == "without-") { nega=true; longo.erase(0, 8); }
    else if(longo.substr(0, 5) == "with-")   { nega=false; longo.erase(0, 5); }
    
NegaDone:
    unsigned p = longo.find('=');
    if(p == longo.npos)p = longo.find(':');
    string option;
    if(p != longo.npos) {option=longo.substr(p+1);longo.erase(p);}
    argh_aliasmap::const_iterator i = aliases->find(longo);
    if(i == aliases->end())
    {
    	if(nega) { nega=false; longo=s+2; goto NegaDone; }
        fprintf(stderr, "%s: unrecognized option: `%s'", A0, s);
        if(negafail)fprintf(stderr, " (can only negate boolean switches)");
        fprintf(stderr, "\n");
        return ParseError();
    }
    if(nega)
    {
	    argh_atypemap::const_iterator j = argtypes->find(i->second);
	    if(j != argtypes->end() && j->second.type != 0)
	    {
	    	negafail = true;
	    	nega = false;
	    	longo = s+2;
	    	goto NegaDone;
	    }
	    polarity = false;
    }
    longobuf = "";
    if(option.size()) { longobuf += '='; longobuf += option; }
    return ShortOpt(i->second, longobuf.c_str());
}

#include <vector>
using std::vector;

void ParamHandler::ListOptions()
{
    argh_atypemap::const_iterator i;
    argh_aliasmap::const_iterator j;
    argh_descsmap::const_iterator k;
    
    vector<pair<string,string> > hdrs;
    unsigned widest=0;

    for(j=aliases->begin(); j!=aliases->end(); ++j)
        if(argtypes->find(j->second) == argtypes->end())
        {
            k = descs->find(j->second);
            if(k == descs->end())continue;

            string s = "    --";
            s += j->first;
            if(k->second.second.size())
            {
                s += ' ';
                s += k->second.second;
            }
            if(s.size() > widest)widest = s.size();
            hdrs.push_back(pair<string,string> (s, k->second.first));
        }
    
    for(i=argtypes->begin(); i!=argtypes->end(); ++i)
    {
        k = descs->find(i->first);
        if(k == descs->end())continue;
        
        string s = "-";
        s += i->first;
        
        for(j=aliases->begin(); j!=aliases->end(); ++j)
            if(j->second == i->first)
            {
                s += ", --";
                s += j->first;
            }
        if(k->second.second.size())
        {
            s += ' ';
            s += k->second.second;
        }
        if(s.size() > widest)widest = s.size();
        hdrs.push_back(pair<string,string> (s, k->second.first));
    }
    
    for(unsigned a=0; a<hdrs.size(); ++a)
    	PrintOpt(widest, hdrs[a].first, hdrs[a].second);
}

ParamHandler::Reference ParamHandler::AddLong(const string &longname, int alias)
{
    (*aliases)[longname]=alias;
    return MakeRef(alias);
}

ParamHandler::Reference ParamHandler::AddDesc(int c, const string &s, const string &param)
{
    (*descs)[c] = pair<string,string>(s,param);
    return MakeRef(c);
}

ParamHandler::Reference ParamHandler::AddBool(int c)
{
    (*argtypes)[c].type=0;
    return MakeRef(c);
}

ParamHandler::Reference ParamHandler::AddInt(int c, int min, int max)
{
    ArgInfo&a=(*argtypes)[c];a.type=1;a.min=min;a.max=max;
    return MakeRef(c);
}

ParamHandler::Reference ParamHandler::AddFloat(int c, double min, double max)
{
    ArgInfo&a=(*argtypes)[c];a.type=2;a.fmin=min;a.fmax=max;
    return MakeRef(c);
}

ParamHandler::Reference ParamHandler::AddString(int c, unsigned min, unsigned max)
{
    ArgInfo&a=(*argtypes)[c];a.type=3;a.min=min;a.maxl=max;
    return MakeRef(c);
}

void ParamHandler::StartParse(int ac, const char *const *av, int firstarg)
{
    argc=ac;argv=av;
    argpos=firstarg-1;opts=true;
    longo="";shortpointer="";
    polarity=true;
    error=false;

    A0 = argv[0];
    for(const char *q; (q = strchr(A0, '/')); )A0 = q+1;
}

#include <cstdio>
using std::printf;
void ParamHandler::PrintOpt(unsigned space, const string &opt, const string &desc)
{
	printf("    %-*s  ", space, opt.c_str());
	
	bool needeol = true;
	unsigned a = 0;
	while(a < desc.size())
	{
	    unsigned b = desc.find('\n', a);
	    if(!needeol){printf("%*s", space+6, "");needeol=true;}
	    if(b == desc.npos) { printf("%s", desc.c_str()+a); break; }
	    printf("%s", desc.substr(a, b-a).c_str());
	    printf("\n"); needeol = false;
	    a = b+1;
    }
    if(needeol)printf("\n");
}

#ifdef NEED_ARGH_H
#ifdef dirr3_argh_hh
#include "argh2.h"
#else
#include "argh.h"
#endif

static ParamHandler *Cph = NULL;
void argh_init(void)
{
	Cph = new ParamHandler();
}
void argh_done(void)
{
	delete Cph;
	Cph = NULL;
}
void argh_add_long(const char *longname, int alias)
{
	Cph->AddLong(longname, alias);
}
void argh_add_bool(int c)
{
	Cph->AddBool(c);
}
void argh_add_int(int c, int min, int max)
{
	Cph->AddInt(c, min, max);
}
void argh_add_float(int c, double min, double max)
{
	Cph->AddFloat(c, min, max);
}
void argh_add_string(int c, unsigned minlen, unsigned maxlen)
{
	Cph->AddString(c, minlen, maxlen);
}
void argh_add_desc(int c, const char *s, const char *optparam)
{
	Cph->AddDesc(c, s, optparam?optparam:"");
}
void argh_start_parse(int argc, const char *const *argv)
{
	Cph->StartParse(argc, argv);
}
int argh_get_bool(void)
{
	return Cph->GetBool();
}
int argh_get_int(void)
{
	return Cph->GetInt();
}
double argh_get_float(void)
{
	return Cph->GetFloat();
}
const char *argh_get_string(void)
{
	return Cph->GetString().c_str();
}
int argh_ok(void)
{
	return Cph->ok();
}
int argh_get_param(void)
{
	return Cph->GetParam();
}
void argh_list_options(void)
{
	Cph->ListOptions();
}
#endif
